from collections import namedtuple

from dagster import check

DEFAULT_CONTEXT_NAME = 'default'


# lifted from https://bit.ly/2HcQAuv
class Context(namedtuple('ContextData', 'name config')):
    def __new__(cls, name=None, config=None):
        return super(Context, cls).__new__(
            cls,
            check.opt_str_param(name, 'name', DEFAULT_CONTEXT_NAME),
            config,
        )


class Solid(namedtuple('Solid', 'config')):
    def __new__(cls, config):
        return super(Solid, cls).__new__(cls, config)


class Environment(namedtuple('EnvironmentData', 'context solids expectations execution')):
    def __new__(cls, solids=None, context=None, expectations=None, execution=None):
        check.opt_inst_param(context, 'context', Context)
        check.opt_inst_param(expectations, 'expectations', Expectations)
        check.opt_inst_param(execution, 'execution', Execution)

        if context is None:
            context = Context()

        if expectations is None:
            expectations = Expectations(evaluate=True)

        if execution is None:
            execution = Execution(serialize_intermediates=False)

        return super(Environment, cls).__new__(
            cls,
            context=context,
            solids=check.opt_dict_param(solids, 'solids', key_type=str, value_type=Solid),
            expectations=expectations,
            execution=execution,
        )


class Expectations(namedtuple('ExpectationsData', 'evaluate')):
    def __new__(cls, evaluate):
        return super(Expectations, cls).__new__(
            cls,
            evaluate=check.bool_param(evaluate, 'evaluate'),
        )


class Execution(namedtuple('ExecutionData', 'serialize_intermediates')):
    def __new__(cls, serialize_intermediates):

        return super(Execution, cls).__new__(
            cls,
            check.bool_param(serialize_intermediates, 'serialize_intermediates'),
        )



from dagster import (
    ConfigDefinition,
    DependencyDefinition,
    InputDefinition,
    OutputDefinition,
    PipelineDefinition,
    SolidInstance,
    define_stub_solid,
    execute_pipeline,
    lambda_solid,
    solid,
    types,
)


@solid(
    config_def=ConfigDefinition(types.Int),
    outputs=[OutputDefinition(types.Int)],
)
def load_number(info):
    return info.config


@lambda_solid(
    inputs=[
        InputDefinition('num1', types.Int),
        InputDefinition('num2', types.Int),
    ],
    output=OutputDefinition(types.Int),
)
def adder(num1, num2):
    return num1 + num2


@lambda_solid(
    inputs=[
        InputDefinition('num1', types.Int),
        InputDefinition('num2', types.Int),
    ],
    output=OutputDefinition(types.Int),
)
def multer(num1, num2):
    return num1 * num2


def define_part_fourteen_step_one_pipeline():
    # (a + b) * (c + d)

    return PipelineDefinition(
        name='part_fourteen_step_one_pipeline',
        solids=[load_number, adder, multer],
        dependencies={
            SolidInstance(load_number.name, 'a'): {},
            SolidInstance(load_number.name, 'b'): {},
            SolidInstance(load_number.name, 'c'): {},
            SolidInstance(load_number.name, 'd'): {},
            SolidInstance(adder.name, 'a_plus_b'): {
                'num1': DependencyDefinition('a'),
                'num2': DependencyDefinition('b'),
            },
            SolidInstance(adder.name, 'c_plus_d'): {
                'num1': DependencyDefinition('c'),
                'num2': DependencyDefinition('d'),
            },
            SolidInstance(multer.name, 'final'): {
                'num1': DependencyDefinition('a_plus_b'),
                'num2': DependencyDefinition('c_plus_d'),
            },
        },
    )


def test_only_final():
    pipeline = PipelineDefinition.create_single_solid_pipeline(
        define_part_fourteen_step_one_pipeline(),
        'final',
        injected_solids={
            'final': {
                'num1': define_stub_solid('stub_a', 3),
                'num2': define_stub_solid('stub_b', 4),
            }
        }
    )

    result = execute_pipeline(pipeline)

    assert result.success
    assert len(result.result_list) == 3
    assert result.result_for_solid('stub_a').transformed_value() == 3
    assert result.result_for_solid('stub_b').transformed_value() == 4
    assert result.result_for_solid('final').transformed_value() == 12


def test_a_plus_b_final_subdag():
    pipeline = PipelineDefinition.create_sub_pipeline(
        define_part_fourteen_step_one_pipeline(),
        ['a_plus_b', 'final'],
        ['final'],
        injected_solids={
            'a_plus_b': {
                'num1': define_stub_solid('stub_a', 2),
                'num2': define_stub_solid('stub_b', 4),
            },
            'final': {
                'num2': define_stub_solid('stub_c_plus_d', 6),
            }
        },
    )

    pipeline_result = execute_pipeline(pipeline)

    assert pipeline_result.result_for_solid('a_plus_b').transformed_value() == 6
    assert pipeline_result.result_for_solid('final').transformed_value() == 36



from dagster import (
    PipelineDefinition,
    check,
    types,
)

from dagster.core.config_types import (
    EnvironmentConfigType,
    is_environment_context_field_optional,
)


def scaffold_pipeline_config(pipeline_def, skip_optional=True):
    check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition)
    check.bool_param(skip_optional, 'skip_optional')

    env_config_type = EnvironmentConfigType(pipeline_def)

    env_dict = {}

    for env_field_name, env_field in env_config_type.field_dict.items():
        if skip_optional and env_field.is_optional:
            continue

        # unfortunately we have to treat this special for now
        if env_field_name == 'context':
            if skip_optional and is_environment_context_field_optional(pipeline_def):
                continue

        env_dict[env_field_name] = scaffold_type(env_field.dagster_type, skip_optional)

    return env_dict


def scaffold_type(config_type, skip_optional=True):
    check.inst_param(config_type, 'config_type', types.DagsterType)
    check.bool_param(skip_optional, 'skip_optional')

    if isinstance(config_type, types.DagsterCompositeType):
        default_dict = {}
        for field_name, field in config_type.field_dict.items():
            if skip_optional and field.is_optional:
                continue

            default_dict[field_name] = scaffold_type(field.dagster_type, skip_optional)
        return default_dict
    elif config_type is types.Dict:
        return {}
    elif config_type is types.Any:
        return 'AnyType'
    elif isinstance(config_type, types.DagsterScalarType):
        defaults = {
            types.String: '',
            types.Path: 'path/to/something',
            types.Int: 0,
            types.Bool: True,
        }

        return defaults[config_type]
    else:
        check.failed('Do not know how to scaffold {type_name}'.format(type_name=config_type.name))



import re
from dagster import check

from .config import (
    Context,
    Environment,
    Execution,
    Expectations,
    Solid,
)

from .definitions import (
    Field,
    PipelineContextDefinition,
    PipelineDefinition,
)

from .types import (
    Bool,
    DagsterCompositeType,
    DagsterEvaluateValueError,
    DagsterType,
    process_incoming_composite_value,
)


class HasUserConfig:
    def __init__(self):
        check.inst(self, DagsterCompositeType, 'HasUserConfig must be mixined on Composite')
        check.invariant(
            'config' in self.field_dict,  # pylint: disable=E1101
            'HasUserConfig must have "config" field',
        )

    @property
    def user_config_field(self):
        return self.field_dict['config']  # pylint: disable=E1101


def load_environment(pipeline_def, environment_dict):
    check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition)
    check.dict_param(environment_dict, 'environment_dict')

    env_type = EnvironmentConfigType(pipeline_def)
    return env_type.evaluate_value(environment_dict)


def define_possibly_optional_field(config_type, is_optional):
    check.inst_param(config_type, 'config_type', DagsterType)
    check.bool_param(is_optional, 'is_optional')

    return Field(
        config_type,
        is_optional=True,
        default_value=lambda: config_type.evaluate_value(None),
    ) if is_optional else Field(config_type)


class SpecificContextConfig(DagsterCompositeType, HasUserConfig):
    def __init__(self, name, config_type):
        check.str_param(name, 'name')
        config_field = define_possibly_optional_field(config_type, all_optional_type(config_type))
        super(SpecificContextConfig, self).__init__(name, {'config': config_field})

    def evaluate_value(self, value):
        config_output = process_incoming_composite_value(self, value, lambda val: val)
        return config_output['config']


def define_specific_context_field(pipeline_name, context_name, context_def, is_optional):
    check.str_param(pipeline_name, 'pipeline_name')
    check.str_param(context_name, 'context_name')
    check.inst_param(context_def, 'context_def', PipelineContextDefinition)
    check.bool_param(is_optional, 'is_optional')

    specific_context_config_type = SpecificContextConfig(
        '{pipeline_name}.ContextDefinitionConfig.{context_name}'.format(
            pipeline_name=pipeline_name,
            context_name=camelcase(context_name),
        ),
        context_def.config_def.config_type,
    )

    return define_possibly_optional_field(specific_context_config_type, is_optional)


def single_item(ddict):
    check.dict_param(ddict, 'ddict')
    check.param_invariant(len(ddict) == 1, 'ddict')
    return list(ddict.items())[0]


class ContextConfigType(DagsterCompositeType):
    def __init__(self, pipeline_name, context_definitions):
        check.str_param(pipeline_name, 'pipeline_name')
        check.dict_param(
            context_definitions,
            'context_definitions',
            key_type=str,
            value_type=PipelineContextDefinition,
        )

        full_type_name = '{pipeline_name}.ContextConfig'.format(pipeline_name=pipeline_name)

        field_dict = {}
        for context_name, context_definition in context_definitions.items():

            is_optional = True if len(context_definitions) > 1 else all_optional_type(
                context_definition.config_def.config_type,
            )

            field_dict[context_name] = define_specific_context_field(
                pipeline_name,
                context_name,
                context_definition,
                is_optional,
            )

        super(ContextConfigType, self).__init__(
            full_type_name,
            field_dict,
            'A configuration dictionary with typed fields',
        )

    def evaluate_value(self, value):
        if isinstance(value, Context):
            return value

        if value is not None and not isinstance(value, dict):
            raise DagsterEvaluateValueError('Incoming value for composite must be None or dict')

        if not value:
            if 'default' not in self.field_dict and len(self.field_dict) > 1:
                raise DagsterEvaluateValueError(
                    'More than one context defined. Must provide one in config'
                )

            # if default is defined use that otherwise use the single context name
            single_context_name, single_context_field = (
                'default',
                self.field_dict['default'],
            ) if 'default' in self.field_dict else single_item(self.field_dict)

            if single_context_field.is_optional and single_context_field.default_provided:
                return Context(single_context_name, single_context_field.default_value)

            raise DagsterEvaluateValueError(
                (
                    'Single context or default context {context_name} defined is not optional '
                    'or default value is not provided. '
                    'Must specify in config'
                ).format(context_name=single_context_name)
            )

        if len(value) > 1:
            specified_contexts = sorted(list(value.keys()))
            available_contexts = sorted(list(self.field_dict.keys()))
            raise DagsterEvaluateValueError(
                (
                    'You can only specify a single context. You specified {specified_contexts}. '
                    'The available contexts are {available_contexts}'
                ).format(
                    specified_contexts=specified_contexts,
                    available_contexts=available_contexts,
                )
            )

        context_name, context_config_value = single_item(value)

        parent_type = self.field_dict[context_name].dagster_type
        config_type = parent_type.field_dict['config'].dagster_type
        processed_value = config_type.evaluate_value(permissive_idx(context_config_value, 'config'))
        return Context(context_name, processed_value)


def permissive_idx(ddict, key):
    check.opt_dict_param(ddict, 'ddict')
    check.str_param(key, 'key')
    if ddict is None:
        return None
    return ddict.get(key)


class SolidConfigType(DagsterCompositeType, HasUserConfig):
    def __init__(self, name, config_type):
        check.str_param(name, 'name')
        super(SolidConfigType, self).__init__(
            name,
            {
                'config': define_possibly_optional_field(
                    config_type,
                    all_optional_type(config_type),
                ),
            },
        )

    def evaluate_value(self, value):
        if isinstance(value, Solid):
            return value

        return process_incoming_composite_value(self, value, lambda val: Solid(**val))

    @property
    def user_config_field(self):
        return self.field_dict['config']


def define_environment_field(field_type):
    check.inst_param(field_type, 'field_type', DagsterType)
    return define_possibly_optional_field(field_type, all_optional_type(field_type))


def has_all_optional_default_context(pipeline_def):
    check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition)
    return 'default' in pipeline_def.context_definitions and all_optional_type(
        pipeline_def.context_definitions['default'].config_def.config_type
    )


def is_environment_context_field_optional(pipeline_def):
    check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition)
    if len(pipeline_def.context_definitions) > 1:
        return has_all_optional_default_context(pipeline_def)
    else:
        _, single_context_def = single_item(pipeline_def.context_definitions)

        return all_optional_type(single_context_def.config_def.config_type)


class EnvironmentConfigType(DagsterCompositeType):
    def __init__(self, pipeline_def):
        check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition)

        pipeline_name = camelcase(pipeline_def.name)

        context_field = define_possibly_optional_field(
            ContextConfigType(pipeline_name, pipeline_def.context_definitions),
            is_environment_context_field_optional(pipeline_def),
        )

        solids_field = define_environment_field(
            SolidDictionaryType(
                '{pipeline_name}.SolidsConfigDictionary'.format(pipeline_name=pipeline_name),
                pipeline_def,
            )
        )

        expectations_field = define_environment_field(
            ExpectationsConfigType(
                '{pipeline_name}.ExpectationsConfig'.format(pipeline_name=pipeline_name)
            )
        )

        execution_field = define_environment_field(
            ExecutionConfigType(
                '{pipeline_name}.ExecutionConfig'.format(pipeline_name=pipeline_name)
            )
        )

        super(EnvironmentConfigType, self).__init__(
            '{pipeline_name}.Environment'.format(pipeline_name=pipeline_name),
            fields={
                'context': context_field,
                'solids': solids_field,
                'expectations': expectations_field,
                'execution': execution_field,
            },
        )

    def evaluate_value(self, value):
        if isinstance(value, Environment):
            return value

        return process_incoming_composite_value(
            self,
            value,
            lambda val: Environment(**val),
        )


class ExpectationsConfigType(DagsterCompositeType):
    def __init__(self, name):
        super(ExpectationsConfigType, self).__init__(
            name,
            {'evaluate': Field(Bool, is_optional=True, default_value=True)},
        )

    def evaluate_value(self, value):
        if isinstance(value, Expectations):
            return value

        return process_incoming_composite_value(self, value, lambda val: Expectations(**val))


def all_optional_type(dagster_type):
    check.inst_param(dagster_type, 'dagster_type', DagsterType)

    if isinstance(dagster_type, DagsterCompositeType):
        for field in dagster_type.field_dict.values():
            if not field.is_optional:
                return False
    return True


def all_optional_user_config(config_type):
    check.inst_param(config_type, 'config_type', HasUserConfig)
    user_config_field = config_type.field_dict['config']
    return all_optional_type(user_config_field.dagster_type)


class SolidDictionaryType(DagsterCompositeType):
    def __init__(self, name, pipeline_def):
        check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition)

        pipeline_name = camelcase(pipeline_def.name)
        field_dict = {}
        for solid in pipeline_def.solids:
            if solid.definition.config_def:
                solid_name = camelcase(solid.name)
                solid_config_type = SolidConfigType(
                    '{pipeline_name}.{solid_name}.SolidConfig'.format(
                        pipeline_name=pipeline_name,
                        solid_name=solid_name,
                    ),
                    solid.definition.config_def.config_type,
                )
                field_dict[solid.name] = define_possibly_optional_field(
                    solid_config_type,
                    all_optional_user_config(solid_config_type),
                )

        super(SolidDictionaryType, self).__init__(name, field_dict)

    def evaluate_value(self, value):
        return process_incoming_composite_value(self, value, lambda val: val)


class ExecutionConfigType(DagsterCompositeType):
    def __init__(self, name):
        check.str_param(name, 'name')
        super(ExecutionConfigType, self).__init__(
            name,
            {
                'serialize_intermediates': Field(Bool, is_optional=True, default_value=False),
            },
        )

    def evaluate_value(self, value):
        return process_incoming_composite_value(self, value, lambda val: Execution(**val))


# Adapted from https://github.com/okunishinishi/python-stringcase/blob/master/stringcase.py
def camelcase(string):
    string = re.sub(r'^[\-_\.]', '', str(string))
    if not string:
        return string
    return str(string[0]).upper() + re.sub(
        r'[\-_\.\s]([a-z])',
        lambda matched: str(matched.group(1)).upper(),
        string[1:],
    )



from dagster import (
    OutputDefinition,
    PipelineDefinition,
    execute_pipeline,
    lambda_solid,
    types,
)


@lambda_solid
def return_one():
    return 1


class Some(object):
    def __init__(self, value):
        self.value = value


SomeType = types.PythonObjectType('SomeType', Some)


def test_serialize_one_int():

    pipeline_def = PipelineDefinition(solids=[return_one])
    result = execute_pipeline(pipeline_def, {'execution': {'serialize_intermediates': True}})
    assert result.success
    path = '/tmp/dagster/runs/{run_id}/return_one/outputs/result'.format(run_id=result.run_id)
    value = types.Any.deserialize_value(path)
    assert value == 1


@lambda_solid(output=OutputDefinition(SomeType))
def return_some_type():
    return Some('foobar')


def test_serialize_custom_type():

    pipeline_def = PipelineDefinition(solids=[return_some_type])
    result = execute_pipeline(pipeline_def, {'execution': {'serialize_intermediates': True}})
    assert result.success
    path = '/tmp/dagster/runs/{run_id}/return_some_type/outputs/result'.format(run_id=result.run_id)
    some = SomeType.deserialize_value(path)
    assert some.value == 'foobar'



from dagster import RepositoryDefinition

from dagster.dagster_examples.tutorials import (
    test_intro_tutorial_part_one,
    test_intro_tutorial_part_two,
    test_intro_tutorial_part_three,
    test_intro_tutorial_part_four,
    test_intro_tutorial_part_five,
    test_intro_tutorial_part_ten,
    test_intro_tutorial_part_eleven,
    test_intro_tutorial_part_twelve,
    test_intro_tutorial_part_thirteen,
    test_intro_tutorial_part_fourteen,
)

from dagster.dagster_examples.tutorials.test_intro_tutorial_part_six import (
    test_intro_tutorial_part_six
)

from dagster.dagster_examples.tutorials.test_intro_tutorial_part_seven import (
    test_intro_tutorial_part_seven
)

from dagster.dagster_examples.tutorials.test_intro_tutorial_part_eight import (
    test_intro_tutorial_part_eight
)

from dagster.dagster_examples.tutorials.test_intro_tutorial_part_nine import (
    test_intro_tutorial_part_nine
)


def define_repository():
    return RepositoryDefinition(
        name='tutorial_repository',
        pipeline_dict={
            'part_one_pipeline':
            test_intro_tutorial_part_one.define_pipeline,
            'part_two_pipeline':
            test_intro_tutorial_part_two.define_pipeline,
            'part_three_pipeline':
            test_intro_tutorial_part_three.define_pipeline,
            'part_four_pipeline':
            test_intro_tutorial_part_four.define_pipeline,
            'part_five_step_one_pipeline':
            test_intro_tutorial_part_five.define_step_one_pipeline,
            'part_five_step_two_pipeline':
            test_intro_tutorial_part_five.define_step_two_pipeline,
            'part_five_step_three_pipeline':
            test_intro_tutorial_part_five.define_step_three_pipeline,
            'part_six_pipeline':
            test_intro_tutorial_part_six.define_part_six_pipeline,
            'part_seven_pipeline':
            test_intro_tutorial_part_seven.define_part_seven_pipeline,
            'part_eight_step_one_pipeline':
            test_intro_tutorial_part_eight.define_part_eight_step_one_pipeline,
            'part_eight_step_two_pipeline':
            test_intro_tutorial_part_eight.define_part_eight_step_two_pipeline,
            'part_eight_step_three_pipeline':
            test_intro_tutorial_part_eight.define_part_eight_step_three_pipeline,
            'part_nine_step_one_pipeline':
            test_intro_tutorial_part_nine.define_part_nine_step_one_pipeline,
            'part_nine_step_two_pipeline':
            test_intro_tutorial_part_nine.define_part_nine_step_two_pipeline,
            'part_nine_final_pipeline':
            test_intro_tutorial_part_nine.define_part_nine_final_pipeline,
            'part_ten_step_one_pipeline':
            test_intro_tutorial_part_ten.define_part_ten_step_one_pipeline,
            'part_eleven_step_one_pipeline':
            test_intro_tutorial_part_eleven.define_part_eleven_step_one_pipeline,
            'part_eleven_step_two_pipeline':
            test_intro_tutorial_part_eleven.define_part_eleven_step_two_pipeline,
            'part_eleven_step_three_pipeline':
            test_intro_tutorial_part_eleven.define_part_eleven_step_three_pipeline,
            'part_twelve_step_one_pipeline':
            test_intro_tutorial_part_twelve.define_part_twelve_step_one_pipeline,
            'part_twelve_step_two_pipeline':
            test_intro_tutorial_part_twelve.define_part_twelve_step_two_pipeline,
            'part_twelve_step_three_pipeline':
            test_intro_tutorial_part_twelve.define_part_twelve_step_three_pipeline,
            'part_twelve_step_four_pipeline':
            test_intro_tutorial_part_twelve.define_part_twelve_step_four_pipeline,
            'part_thirteen_step_one_pipeline':
            test_intro_tutorial_part_thirteen.define_part_thirteen_step_one_pipeline,
            'part_thirteen_step_two_pipeline':
            test_intro_tutorial_part_thirteen.define_part_thirteen_step_two_pipeline,
            'part_thirteen_step_three_pipeline':
            test_intro_tutorial_part_thirteen.define_part_thirteen_step_three_pipeline,
            'part_fourteen_step_one_pipeline':
            test_intro_tutorial_part_fourteen.define_part_fourteen_step_one_pipeline,
        },
    )



# pylint: disable=W0622,W0614,W0401
from dagster import (
    DependencyDefinition,
    InputDefinition,
    PipelineDefinition,
    execute_pipeline,
    lambda_solid,
)


@lambda_solid
def solid_one():
    return 'foo'


@lambda_solid(inputs=[InputDefinition('arg_one')])
def solid_two(arg_one):
    print(arg_one * 2)


def define_pipeline():
    return PipelineDefinition(
        name='part_two_pipeline',
        solids=[solid_one, solid_two],
        dependencies={
            'solid_two': {
                'arg_one': DependencyDefinition('solid_one'),
            },
        }
    )


def test_tutorial_part_two():
    pipeline_result = execute_pipeline(define_pipeline())
    assert pipeline_result.success
    return pipeline_result


if __name__ == '__main__':
    test_tutorial_part_two()



from dagster import (
    ConfigDefinition,
    DependencyDefinition,
    InputDefinition,
    OutputDefinition,
    PipelineDefinition,
    SolidInstance,
    config,
    execute_pipeline,
    lambda_solid,
    solid,
    types,
)


@solid(config_def=ConfigDefinition(types.Int), outputs=[OutputDefinition(types.Int)])
def load_a(info):
    return info.config


@solid(config_def=ConfigDefinition(types.Int), outputs=[OutputDefinition(types.Int)])
def load_b(info):
    return info.config


@lambda_solid(
    inputs=[
        InputDefinition('a', types.Int),
        InputDefinition('b', types.Int),
    ],
    output=OutputDefinition(types.Int),
)
def a_plus_b(a, b):
    return a + b


def define_part_thirteen_step_one_pipeline():
    return PipelineDefinition(
        name='part_thirteen_step_one_pipeline',
        solids=[load_a, load_b, a_plus_b],
        dependencies={
            'a_plus_b': {
                'a': DependencyDefinition('load_a'),
                'b': DependencyDefinition('load_b'),
            }
        }
    )


def test_part_thirteen_step_one():
    pipeline_result = execute_pipeline(
        define_part_thirteen_step_one_pipeline(),
        {
            'solids': {
                'load_a': {
                    'config': 234,
                },
                'load_b': {
                    'config': 384
                },
            },
        },
    )

    assert pipeline_result.success
    solid_result = pipeline_result.result_for_solid('a_plus_b')
    assert solid_result.transformed_value() == 234 + 384


@solid(
    config_def=ConfigDefinition(types.Int),
    outputs=[OutputDefinition(types.Int)],
)
def load_number(info):
    return info.config


@lambda_solid(
    inputs=[
        InputDefinition('num1', types.Int),
        InputDefinition('num2', types.Int),
    ],
    output=OutputDefinition(types.Int),
)
def adder(num1, num2):
    return num1 + num2


def define_part_thirteen_step_two_pipeline():
    return PipelineDefinition(
        name='part_thirteen_step_two_pipeline',
        solids=[load_number, adder],
        dependencies={
            SolidInstance('load_number', alias='load_a'): {},
            SolidInstance('load_number', alias='load_b'): {},
            SolidInstance('adder', alias='a_plus_b'): {
                'num1': DependencyDefinition('load_a'),
                'num2': DependencyDefinition('load_b'),
            }
        }
    )


def test_part_thirteen_step_two():
    pipeline_result = execute_pipeline(
        define_part_thirteen_step_two_pipeline(),
        {
            'solids': {
                'load_a': {
                    'config': 23,
                },
                'load_b': {
                    'config': 38
                },
            },
        },
    )

    assert pipeline_result.success
    solid_result = pipeline_result.result_for_solid('a_plus_b')
    assert solid_result.transformed_value() == 23 + 38


@lambda_solid(
    inputs=[
        InputDefinition('num1', types.Int),
        InputDefinition('num2', types.Int),
    ],
    output=OutputDefinition(types.Int),
)
def multer(num1, num2):
    return num1 * num2


def define_part_thirteen_step_three_pipeline():
    # (a + b) * (c + d)

    return PipelineDefinition(
        name='part_thirteen_step_three_pipeline',
        solids=[load_number, adder, multer],
        dependencies={
            SolidInstance(load_number.name, 'a'): {},
            SolidInstance(load_number.name, 'b'): {},
            SolidInstance(load_number.name, 'c'): {},
            SolidInstance(load_number.name, 'd'): {},
            SolidInstance(adder.name, 'a_plus_b'): {
                'num1': DependencyDefinition('a'),
                'num2': DependencyDefinition('b'),
            },
            SolidInstance(adder.name, 'c_plus_d'): {
                'num1': DependencyDefinition('c'),
                'num2': DependencyDefinition('d'),
            },
            SolidInstance(multer.name, 'final'): {
                'num1': DependencyDefinition('a_plus_b'),
                'num2': DependencyDefinition('c_plus_d'),
            },
        },
    )


def test_run_whole_pipeline():
    pipeline = define_part_thirteen_step_three_pipeline()
    pipeline_result = execute_pipeline(
        pipeline,
        config.Environment(
            solids={
                'a': config.Solid(2),
                'b': config.Solid(6),
                'c': config.Solid(4),
                'd': config.Solid(8),
            }
        )
    )

    assert pipeline_result.success

    assert pipeline_result.result_for_solid('a_plus_b').transformed_value() == 8
    assert pipeline_result.result_for_solid('c_plus_d').transformed_value() == 12
    assert pipeline_result.result_for_solid('final').transformed_value() == 8 * 12



from collections import namedtuple
import os
import pickle
import tempfile

import pandas as pd

from dagster import (
    ConfigDefinition,
    Field,
    ExecutionContext,
    InputDefinition,
    OutputDefinition,
    Result,
    SolidDefinition,
    check,
    types,
)

DataFrameMeta = namedtuple('DataFrameMeta', 'format path')


class _DataFrameType(types.PythonObjectType):
    def __init__(self):
        super(_DataFrameType, self).__init__(
            name='PandasDataFrame',
            python_type=pd.DataFrame,
            description='''Two-dimensional size-mutable, potentially heterogeneous
    tabular data structure with labeled axes (rows and columns). See http://pandas.pydata.org/''',
        )

    def create_serializable_type_value(self, value, output_dir):
        check.str_param(output_dir, 'output_dir')
        csv_path = os.path.join(output_dir, 'csv')
        value.to_csv(csv_path, index=False)
        df_meta = DataFrameMeta(format='csv', path='csv')
        return types.SerializedTypeValue(name=self.name, value=df_meta._asdict())

    def deserialize_from_type_value(self, type_value, output_dir):
        check.str_param(output_dir, 'output_dir')

        df_meta_dict = type_value.value
        check.inst(df_meta_dict, dict)
        df_meta = DataFrameMeta(**df_meta_dict)

        if df_meta.format == 'csv':
            csv_path = os.path.join(output_dir, df_meta.path)
            return pd.read_csv(csv_path)
        else:
            raise Exception('unsupported')


DataFrame = _DataFrameType()

LoadDataFrameConfigDict = types.ConfigDictionary(
    'LoadDataFrameConfigDict',
    {
        'path': Field(types.Path),
    },
)

WriteDataFrameConfigDict = types.ConfigDictionary(
    'WriteDataFrameConfigDict',
    {
        'path': Field(types.Path),
    },
)


def load_csv_solid(name):
    check.str_param(name, 'name')

    def _t_fn(info, _inputs):
        yield Result(pd.read_csv(info.config['path']))

    return SolidDefinition(
        name=name,
        inputs=[],
        outputs=[OutputDefinition(DataFrame)],
        transform_fn=_t_fn,
        config_def=ConfigDefinition(LoadDataFrameConfigDict),
    )


def to_csv_solid(name):
    def _t_fn(info, inputs):
        inputs['df'].to_csv(info.config['path'], index=False)

    return SolidDefinition(
        name=name,
        inputs=[InputDefinition('df', DataFrame)],
        outputs=[],
        config_def=ConfigDefinition(WriteDataFrameConfigDict),
        transform_fn=_t_fn,
    )


def to_parquet_solid(name):
    def _t_fn(info, inputs):
        inputs['df'].to_parquet(info.config['path'])

    return SolidDefinition(
        name=name,
        inputs=[InputDefinition('df', DataFrame)],
        outputs=[],
        config_def=ConfigDefinition(WriteDataFrameConfigDict),
        transform_fn=_t_fn,
    )



from dagster import (
    PipelineDefinition,
    execute_pipeline,
    lambda_solid,
)


def test_tutorial_part_one():
    pipeline = define_pipeline()

    result = execute_pipeline(pipeline)

    assert result.success
    assert len(result.result_list) == 1
    assert result.result_for_solid('hello_world').transformed_value() is None
    return result


def define_pipeline():
    return PipelineDefinition(name='part_one_pipeline', solids=[hello_world])


@lambda_solid
def hello_world():
    print('hello')


if __name__ == '__main__':
    test_tutorial_part_one()



# pylint: disable=W0622,W0614,W0401
from dagster import (
    DependencyDefinition,
    InputDefinition,
    PipelineDefinition,
    execute_pipeline,
    lambda_solid,
)

#   B
#  / \
# A   D
#  \ /
#   C


@lambda_solid
def solid_a():
    print('a: 1')
    return 1


@lambda_solid(inputs=[InputDefinition('arg_a')])
def solid_b(arg_a):
    print('b: {b}'.format(b=arg_a * 2))
    return arg_a * 2


@lambda_solid(inputs=[InputDefinition('arg_a')])
def solid_c(arg_a):
    print('c: {c}'.format(c=arg_a * 3))
    return arg_a * 3


@lambda_solid(inputs=[InputDefinition('arg_b'), InputDefinition('arg_c')])
def solid_d(arg_b, arg_c):
    print('d: {d}'.format(d=arg_b * arg_c))


def define_pipeline():
    return PipelineDefinition(
        name='part_three_pipeline',
        solids=[solid_d, solid_c, solid_b, solid_a],
        dependencies={
            'solid_b': {
                'arg_a': DependencyDefinition('solid_a'),
            },
            'solid_c': {
                'arg_a': DependencyDefinition('solid_a'),
            },
            'solid_d': {
                'arg_b': DependencyDefinition('solid_b'),
                'arg_c': DependencyDefinition('solid_c'),
            }
        }
    )


def test_tutorial_part_three():
    execute_pipeline(define_pipeline())


if __name__ == '__main__':
    test_tutorial_part_three()
